@@ -83,6 +83,9 @@ def package_file(filename):
             else:
                 time.sleep(.1)
         elif not os.path.exists(filename + '.url'):
+            if os.path.exists(filename + '.cancel'):
+                # file doesn't exist at the other end...
+                abort(404, "File does not exist at remote URL")
             # beaten; file's already downloaded
             log.info('%s appears to be fully downloaded now', filename)
             return static_file(filename, './')
@@ -121,6 +124,19 @@ def download(url, filename):
         with open(local_path, 'wb') as content_file:
             content = urllib.urlopen(url)
 
+            # urlopen will automatically follow redirects, so any non-200 code
+            # is an error
+            http_code = content.getcode()
+            if http_code != 200:
+                log.info('url %s got HTTP %d, cancelling download', url,
+                    http_code)
+                with open(filename + '.cancel', 'w') as cancel_file:
+                    cancel_file.write('HTTP %d' % http_code)
+                # don't attempt to re-download
+                os.remove(filename + '.url')
+                log.info('wrote cancel file %s.cancel', filename)
+                return
+
             # get file size and store it away
             size = int(content.info()['Content-Length'])
             with open(filename + '.size', 'w') as size_file:
@@ -167,11 +183,11 @@ def partial_static_file(filename, root, actual_size):
     log.info('serving %s which is being download', filename)
 
     if not filename.startswith(root):
-        return abort(403, "Access denied.")
+        abort(403, "Access denied.")
     if not os.path.exists(filename) or not os.path.isfile(filename):
-        return abort(404, "File does not exist.")
+        abort(404, "File does not exist.")
     if not os.access(filename, os.R_OK):
-        return abort(403, "You do not have permission to access this file.")
+        abort(403, "You do not have permission to access this file.")
 
     mimetype, encoding = mimetypes.guess_type(filename)
     if mimetype:
@@ -217,7 +233,7 @@ def find_files(directory):
                 # this is a file-that-could-be
                 yield (file[:-4], os.path.relpath(os.path.join(root,
                     file[:-4]), directory))
-            elif file.endswith('.size'):
+            elif file.endswith('.size') or file.endswith('.cancel'):
                 continue
             else:
                 yield (file, os.path.relpath(os.path.join(root, file),
@@ -271,6 +287,17 @@ def locate(pkgname):
         if not os.path.exists(p.name):
             os.mkdir(p.name)
         filenames.append(filename)
+
+        # file has already been fetched
+        if os.path.exists(filename):
+            continue
+
+        # clear any failed previous attempt to fetch - we should re-try just in
+        # case the file (re-)appears
+        if os.path.exists(filename + '.cancel'):
+            os.remove(filename + '.cancel')
+
+        # set up future download of the file
         with open(filename + '.url', 'w') as f:
             f.write(p.download_url)
 
